/* Placed in the public domain 2001 by Sam Trenholme */

/* This is the only code that uses threads */
#include <pthread.h>

/* The MaraDNS includes */
#include "../libs/MaraHash.h"
#include "../MaraDns.h"
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/param.h>
#include "../dns/functions_dns.h"
#include "../parse/functions_parse.h"
#include "../rijndael/rijndael-api-fst.h"
#include "functions_server.h"

pthread_mutex_t big_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t rng_lock = PTHREAD_MUTEX_INITIALIZER;
#ifdef CALL_LOCKS
int read_lock() {
     printf("Calling read_lock()\n");
     return pthread_mutex_lock(&big_lock);
     }
int read_unlock() {
     printf("Calling read_unlock()\n");
     return pthread_mutex_unlock(&big_lock);
     }
int write_lock() {
     printf("Calling write_lock()\n");
     return pthread_mutex_lock(&big_lock);
     }
int write_unlock() {
     printf("Calling write_unlock()\n");
     return pthread_mutex_unlock(&big_lock);
     }
#define read2write_lock() 
#define write2read_lock() 
#else
#define read_lock() pthread_mutex_lock(&big_lock)
#define read_unlock() pthread_mutex_unlock(&big_lock)
#define write_lock() pthread_mutex_lock(&big_lock)
#define write_unlock() pthread_mutex_unlock(&big_lock)
#define read2write_lock() 
#define write2read_lock() 
#endif /* CALL_LOCKS */
#define srng_lock() pthread_mutex_lock(&rng_lock)
#define srng_unlock() pthread_mutex_unlock(&rng_lock)

/* How long records will stay in the cache by default */
#define DEFAULT_TTL 86400

mhash *dnscache = 0;

typedef struct {
    int id;
    int sock;
    struct sockaddr_in client;
    js_string *query;
    } dnsreq; /* DNS request */

/* DNS server that is closer to the answer for a given request */
typedef struct closer {
    unsigned int num_elements; /* Number of elements in the linked list */
    time_t ttd; /* Time for this record to die */
    int datatype; /* This is the RR type of the data (RR_A [in which
                     case, the data is a pointer to a uint32] or RR_NS [which
		     makes the data point to a js_string object]) */
    void *data;
    struct closer *masked; /* If another record with a ttd after the
                              ttd for this record exists, we want to 
			      access that data when this record expires */
    struct closer *next; } closer;

/* The final argument to the functions that add elements to the cache */
#define OVERWRITE 1
#define APPEND 2

/* The variables that the srng (secure random-number generator) uses */
BYTE r_inBlock[17],r_outBlock[17],r_binKey[17];
BYTE r_keyMaterial[320]; /* We may not eventually need this */
keyInstance r_keyInst;
cipherInstance r_cipherInst;
int r_place = 0;
int custodian_mode = 0; /* custodian mode: erase records from the cache
			   every time someone adds a records to the cache */
int cache_max = 0;

#ifdef DEBUG
/* Debug routine that shows an IP in dotted decimal format on the
   standard output.
   Input:  A uint32 ip 
   Output: none
*/
void debug_show_ip(uint32 ip) {
    printf("%d.%d.%d.%d",(ip & 0xff000000) >> 24,
                         (ip & 0x00ff0000) >> 16,
                         (ip & 0x0000ff00) >>  8,
                          ip & 0x000000ff);
    }

/* Debug routine to browse the contents of an assosciate array (dictionary,
   hash table)
   Input: Pointer to hash table
   Output: none
*/

void debug_explore_dict(mhash *dict) {
     /* Lots of cut and paste from mtest.c */
     char strn[256];
     mhash_offset counter,number;
     closer *close;
     rr *dnsrr;

#ifdef INTERACTIVE
     for(;;) {
         printf("s to see the entire hash, v for verbose view, q to resume ");
	 printf("execution of maradns\n");
	 printf("Enter command: ");
	 fgets(strn,200,stdin);
#else
	 *strn = 'v';
#endif /* INTERACTIVE */
	 if(strn[strlen(strn) - 1] == '\n')
	     strn[strlen(strn) - 1] = '\0';
         if(*strn == 's' || *strn == 'v') {
	     printf("Viewing the assosciative array\n");
	     /* Make number 2^hash_bits */
	     number = 1;
	     for(counter = 0; counter < dict->hash_bits; counter++)
                 number *= 2;
             for(counter = 0; counter < number; counter++) {
	         if(dict->hash_table[counter] != 0 &&
		    dict->hash_table[counter]->key != 0 &&
		    js_has_sanity(dict->hash_table[counter]->key)) {
		     printf("Key: ");
		     show_esc_stdout(dict->hash_table[counter]->key);
                     printf("\nValue: ");
		     if(dict->hash_table[counter]->datatype == MARA_DNS_NS) {
		         printf("Name server");
			 if(*strn == 'v') {
			     close = dict->hash_table[counter]->value;
			     printf("\nNum_elements: %d\n", 
			            close->num_elements);
			     while(close != NULL) {
			         printf("Time to die: %d", close->ttd);
			         printf(" Datatype: %d", close->datatype);
			         if(close->datatype == RR_A) {
			             printf(" A Data: ");
				     debug_show_ip(*(uint32 *)(close->data));
				     }
				 else if(close->datatype == RR_NS) {
			             printf(" NS Data: ");
				     show_esc_stdout(
						 (js_string *)(close->data));
				     }
			         printf(" Address: %p",close);
                                 close = close->next;
                                 printf("\n");
				 }
			     }
			 }
                     else if(dict->hash_table[counter]->datatype 
		             == MARA_DNSRR || 
			     dict->hash_table[counter]->datatype
			     == MARA_DNS_NEG) {
			 if(dict->hash_table[counter]->datatype ==
				 MARA_DNSRR)
		             printf("DNS resource record");
			 else
			     printf("Negative RR");
			 if(*strn == 'v') {
			     dnsrr = dict->hash_table[counter]->value;
			     printf("\nExpire: %d",dnsrr->expire);
			     printf(" TTL: %d",dnsrr->ttl);
			     printf(" auth: %d",dnsrr->authoritative);
			     printf(" rr_type: %d",dnsrr->rr_type);
			     printf("\nData: ");
			     show_esc_stdout(dnsrr->data);
                             printf("\n");
			     }
			 }
                     else
		         printf("Non-DNS datatype (why is this here?)");
                     printf("\n\n");
		     }
                 } 	
             }
#ifdef INTERACTIVE
         else if(*strn == 'q') 
	     break;
	 }    
#endif /* INTERACTIVE */
    }               
#endif /* DEBUG */

/* Code that handles the "custodial" maintainence of the cache.
   Basically, if we hit the maximum number of elements in the cache,
   this hits "custodian" mode, where we eliminate eight records at random
   for each record we add to the cache (choose a random spot in the hash,
   erase the next eight records from there [wrapping round as needed])
   Input: None  (global variables used)
   Output: JS_ERROR if something really bad happened, otherwise
           JS_SUCCESS
*/

int custodian() {
    int counter;
    int hsize, hpoint, loop_hpoint;
    closer *cpoint;
    /* Sanity check */
    if(dnscache == 0 || cache_max == 0)
	return JS_ERROR;
    /* If we are within our limits, do nothing */
    if(dnscache->spots < cache_max && custodian_mode <= 0)
        return JS_SUCCESS;	
    /* If we have exceeded the maximum size, go in to "custodian"
       mode, where we destroy one percent of the records in the cache */
    if(dnscache->spots >= cache_max)
	custodian_mode = cache_max / 100;
    /* Otherwise, eight random records get the axe */
    hsize = 1;
    for(counter = 0; counter < dnscache->hash_bits; counter++) 
	hsize *= 2;
    hpoint = random() % hsize; 
    loop_hpoint = hpoint;

    counter = 8;
    /*write_lock(); Not wanted, custodian is only called when the hash
                    is write-locked */
    while(counter > 0) {

	/* Go to the next spot in the hash with data */
	while(dnscache->hash_table[hpoint] == 0) {
	    hpoint++;
	    hpoint = hpoint % hsize;
	    if(hpoint == loop_hpoint)
		return JS_SUCCESS; /* Nothing to remove */
	    }

	/* Destroy the hash element pointed to by the hpoint pointer */
	switch(dnscache->hash_table[hpoint]->datatype) {
	    case MARA_DNSRR:
	    case MARA_DNS_NEG:
		unlink_rr(dnscache->hash_table[hpoint]->value,0);
		break;
            case MARA_DNS_NS:
	        cpoint = dnscache->hash_table[hpoint]->value;
		if(cpoint->ttd != 0) /* Never nuke root server entries */
		    unlink_closer(dnscache->hash_table[hpoint]->value);
            }
	js_destroy(dnscache->hash_table[hpoint]->key);
	js_dealloc(dnscache->hash_table[hpoint]);
	dnscache->hash_table[hpoint] = 0;
	dnscache->spots--;

	counter--;
        }

    /* write_unlock(); Not wanted (see above) */
    custodian_mode--;
    return JS_SUCCESS;

    }

/* A cryptgraphically secure random number with a value between
   0 and 65535
   Input: None
   Output: A number between 0 and 65535
*/
uint16 srng() {
     uint16 ret;
     srng_lock();
     /* If needed, rerun the encryption to create 128 more random bits */
     if(r_place >= 16) {
	 r_inBlock[8] = r_outBlock[12];
         r_inBlock[9] = r_outBlock[13];
         r_inBlock[10] = r_outBlock[14];
         r_inBlock[11] = r_outBlock[15];
         time((time_t *)&r_inBlock[0]);
         blockEncrypt(&r_cipherInst,&r_keyInst,r_inBlock,128,r_outBlock);
	 r_place = 0;
         }
     ret = ((r_outBlock[r_place] & 0xff) << 8) | 
	    (r_outBlock[r_place + 1] & 0xff);
     r_place += 2;
     srng_unlock(); 
     return ret;
     }

/* Destory a given linked list of "closer" records.
   Input:  Pointer to linked list we wish to destroy
   Output: JS_ERROR on error, JS_SUCCESS on success
*/

int unlink_closer(closer *fatma) {
   
    closer *remember;

    while(fatma != NULL) {
        switch(fatma->datatype) {
	    case RR_NS:
	        if(fatma->data != NULL)
	            js_destroy(fatma->data);
		break;
            default:
	        if(fatma->data != NULL)
	            js_dealloc(fatma->data);
                break;
            }
	remember = fatma;
        fatma = fatma->next;
	js_dealloc(remember);
	}
    
    return JS_SUCCESS;
    }

/* Destory a given linked list of resource records.
   Input:  Pointer to linked list we wish to destroy
   Output: JS_ERROR on error, JS_SUCCESS on success
*/

int unlink_rr(rr *fatma,int depth) {
   
    rr *remember;
    
    if(depth > 32)
	return JS_ERROR; /* Maximum allowed depth is 10 */

    while(fatma != NULL) {
	if(fatma->ip != NULL)
	    unlink_rr(fatma->ip,depth + 1);
	/* We do not destroy fatma->query because that js string object
	   has already been destroyed */
	js_destroy(fatma->data);
	remember = fatma;
        fatma = fatma->next;
	js_dealloc(remember);
	}
    
    return JS_SUCCESS;
    }

/* Determine if a given substring of one long string is the same value as
   another string.
   Input: Pointer to first long js_string obejct, offset from which we
          begin comparison, length of subsction of string we compare,
	  Pointer to second short js_string object 
   Ouput: JS_ERROR on fatal error, 0 if they are different, 1 if they
          are the same 
*/

int substring_issame(js_string *longjs, int offset, int length, 
                     js_string *shortjs) {
    js_string *compare;
    int ret;
    /* Sanity checks */
    if(longjs->unit_size != 1 || shortjs->unit_size != 1)
        return JS_ERROR;
    if(offset < 0 || length < 1)
        return JS_ERROR;
    /* Create the temporary string used for comparision purposes */
    if((compare = js_create(length + 1,1)) == 0)  
        return JS_ERROR;
    if(js_substr(longjs,compare,offset,length) == JS_ERROR) {
        js_destroy(compare);
	return JS_ERROR;
        }
    ret = js_issame(shortjs,compare);
    js_destroy(compare);
    return ret;
    }

/* Determine if a given substring of one long string is the same value as
   another string, taking in to account that this may be case-insensitive,
   and taking in to account that a query for "RR_ANY" can return any record
   type.
   Input: Pointer to first long js_string obejct, offset from which we
          begin comparison, length of subsction of string we compare,
	  Pointer to second short js_string object 
   Ouput: JS_ERROR on fatal error, 0 if they are different, 1 if they
          are the same 
*/

int substring_issame_case(js_string *longjs, int offset, int length, 
                          js_string *shortjs) {
    js_string *compare, *lower;
    int ret, case_folded1, case_folded2, counter, qtype_long,
    qtype_short;
    lower = 0;
    /* Sanity checks */
    if(longjs->unit_size != 1 || shortjs->unit_size != 1)
        return JS_ERROR;
    if(offset < 0 || length < 1)
        return JS_ERROR;
    /* Create the temporary string used for comparision purposes */
    if((compare = js_create(length + 1,1)) == 0)  
        return JS_ERROR;
    if(js_substr(longjs,compare,offset,length) == JS_ERROR) {
        js_destroy(compare);
	return JS_ERROR;
        }
    /* Get the query types of the short and the compare strings */
    qtype_long = get_rtype(compare);
    qtype_short = get_rtype(shortjs);
    /* Yet another sanity check */
    if(qtype_long == JS_ERROR || qtype_short == JS_ERROR) {
        js_destroy(compare);
	return JS_ERROR;
        }
    /* If the short js is an ANY query, it will match whatever query
       type the long js has */
    if(qtype_short == RR_ANY) {
        if(change_rtype(shortjs,qtype_long) == JS_ERROR) {
	    js_destroy(compare);
	    return JS_ERROR;
	    }
        }
    
    /* Do the first actual comparison */
    ret = js_issame(shortjs,compare);
    /* Sanity check */
    if(ret == JS_ERROR) {
	js_destroy(compare);
	change_rtype(shortjs,qtype_short);
	return JS_ERROR;
        }
    /* If they do not match, pweform a case-insensitive search */
    if(ret == 0) {
        if((lower = js_create(shortjs->unit_count + 1,1)) == 0) {
	    js_destroy(compare);
	    change_rtype(shortjs,qtype_short);
	    return JS_ERROR;
	    }
	ret = JS_ERROR; /* Return appropraite error if we go to cleanup */
	if(js_copy(shortjs,lower) == JS_ERROR)
	    goto cleanup;
	case_folded1 = fold_case(lower);
	case_folded2 = fold_case(compare);
	/* If either string had folded case 
	   (one or more upper case letters) */
	if(case_folded1 == JS_SUCCESS || case_folded2 == JS_SUCCESS) {
	    /* See if the case-insensitive forms are the same.  If not,
	       destroy strings and return 0. */
            ret = js_issame(lower,compare);
	    if(ret == JS_ERROR)
		goto cleanup;
	    if(ret == 0) 
	        goto cleanup;
	    /* OK, they match when both are lowercase.  Now fold the case 
	       of the strings we sent to the routine */
	    /* Folding the case of the short string is simple (and, yes,
	       we could have done this above, but this is simpler to read) */
	    if(case_folded1 == JS_SUCCESS)
		fold_case(shortjs);
	    /* Folding the case of a substring of the longer string
	       is not-so-simple */
	    if(case_folded2 == JS_SUCCESS) {
                counter = offset;
		while(counter + 2 < offset + length) {
		    if(*(longjs->string + counter) >= 'A' &&
		       *(longjs->string + counter) <= 'Z') {
                        *(longjs->string + counter) += 32;
		        }
		    counter++;
		    }
	        }
	    }
        }
    cleanup:
        if(lower != 0)
	    js_destroy(lower);
        js_destroy(compare);
	change_rtype(shortjs,qtype_short);
        return ret;
    }

/* This is how we are both case-insensitive and 
   case-perserving with domain names.  If the case
   of the answer differs from the case of the
   question, we make what we add to the cache
   lower-case (case-insensitive) */

int check_case_of_answer(js_string *uindata, js_string *query, int offset) {
    js_string *copy;
    int result;
    if((copy = js_create(query->unit_count + 3,1)) == 0)
	return JS_ERROR;
    if(js_copy(query,copy) == JS_ERROR) {
	js_destroy(copy);
	return JS_ERROR;
        }
    /* Since it is the NS records we need to change the case of, make
       copy a NS query */
    if(change_rtype(copy,RR_NS) == JS_ERROR) {
	js_destroy(copy);
        return JS_ERROR;
        }

    result = 0;
    do {
        /* We change the case of the query as appropriate */
        result = substring_issame_case(uindata,offset,
		    dlabel_length(uindata,offset) + 2,
		    copy);
#ifdef DEBUG
	printf("Query: ");
	show_esc_stdout(copy);
	printf(" Result: %d\n",result);
#endif
        } while(bobbit_label(copy) > 0 && result <= 0);

    js_destroy(copy);
    return JS_SUCCESS;
    }

/* Add a substring of a js_string object to the cache hash.
   Input: Pointer to hash to add element to, Pointer to long string, 
          offset of substr, length of substr, ttl of query, pointer to 
	  query, whether to put new element (1) or to add to end of 
	  existing object (2)
   output: JS_ERROR on error, JS_SUCCESS on success
*/

substring_add_rr(mhash *hashp, js_string *longjs, int offset, int length,
                 uint32 ttl, js_string *query, int action, int datatype) {
    js_string *sub;
    int ret, qtype;
    uint32 ttd;
    /* Sanity checks */
    if(longjs->unit_size != 1 || query->unit_size != 1)
        return JS_ERROR; 
    if(offset < 0 || length < 1)
        return JS_ERROR;
    if(action < 1 || action > 2)
        return JS_ERROR;

    /* Create and set the value of the substr that we will put the value in */
    if((sub = js_create(length + 1,1)) == 0)
        return JS_ERROR;
    if(js_substr(longjs,sub,offset,length) == JS_ERROR) {
        js_destroy(sub);
	return JS_ERROR;
	}

    /* Make the ttl the time this record dies */
    ttd = time(0) + ttl;
    write_lock();
    ret = mhash_put_data(hashp,query,sub,ttl,0,ttd,datatype,255,action);
    custodian();
    write_unlock(); 
    js_destroy(sub);
    return ret;
    }

/* Determine if a host label with is a substring of a long js
   object is in bailiwick.
   Input: String with host name, offset of beginning of host name,
          bailiwick we are allowed to be in
   Output: JS_ERROR (-1) on fatal error, 0 if host name is out
           of bailiwick, 1 if host name is in bailiwick 
*/

int offset_bailiwick(js_string *js, int offset, js_string *bailiwick) {
    js_string *sub;
    int len, ret, len_save;

    /* Sanity check */
    if(bailiwick->unit_size != 1)
        return JS_ERROR;

    /* Determine how long the domain label is */
    len = dlabel_length(js,offset);
    if(len == JS_ERROR)
        return JS_ERROR;

    /* Make a string which holds just the domain name label */
    if((sub = js_create(len + 1,1)) == 0)
        return JS_ERROR;
    if(js_substr(js,sub,offset,len) == JS_ERROR) {
        js_destroy(sub);
	return JS_ERROR;
	}

    /* Just in case Bailiwick is longer than the dlabel at the top of the
       Bailiwick, truncate the bailiwick */
    len_save = bailiwick->unit_count;
    len = dlabel_length(bailiwick,0);
    if(len == JS_ERROR) {
        js_destroy(sub);
	return JS_ERROR;
	}
    bailiwick->unit_count = len;
    /* Run in_bailiwick with the substring */
    ret = in_bailiwick(sub,bailiwick);
    bailiwick->unit_count = len_save;
    js_destroy(sub);
    return ret;
    }

/* Determine if a given host name is in the bailiwick of another 
   domain name.
   Input: Host name we are querying, bailiwick we are allowed to be in
   Output: JS_ERROR (-1) on fatal error, 0 if host name is out
           of bailiwick, 1 if host name is in bailiwick 
*/

int in_bailiwick(js_string *host, js_string *bailiwick) {
    js_string *get;
    int result;
    if((get = js_create(host->unit_count + 3,1)) == 0)
        return JS_ERROR;
    if(js_copy(host,get) == JS_ERROR)
        return JS_ERROR;
    get->encoding = bailiwick->encoding;
    result = 0; /* Out of bailiwick */
    if(js_issame(get,bailiwick) == 1)
        return 1; /* In bailiwick */
    while(result == 0 && get->unit_count > bailiwick->unit_count) {
        if(bobbit_label(get) == JS_ERROR)
	    return JS_ERROR;
	if(js_issame(get,bailiwick) == 1)
	    return 1; /* In bailiwick */
        }
    return result; /* Out of bailiwick */
    }


/* Add a name to the dns cache
   Input: Pointer to hash, host name, host ttl, authoritative flag, 
          expire, datatype (whether this is a positive or negative answer),
	  rtype of data (255 means it uses the rtype that query has), and
	  action to perform (1: Destory any element that already exists
	  at the hash spot in question, if applicable.  2: Append to any
	  element that is already exists at the hash spot in question)
   Output: JS_ERROR on error, JS_SUCCESS on success
*/

int mhash_put_data(mhash *hash, js_string *query, js_string *value, uint32 ttl,
                   uint32 authoritative, uint32 expire, int datatype,
		   int rtype, int action) {

    rr *data, *point;
    js_string *new, *new_query;
    int ret, rrtype;
    mhash_e spot_data;

    /* Create a structure for putting the rr data in */
    if((data = js_alloc(1,sizeof(rr))) == 0)
        return JS_ERROR;

    /* First, clear out all the fields */
    data->expire = data->ttl = data->authoritative = data->rr_type = 0;
    data->next = data->ip = 0;
    data->query = data->data = 0;
    data->seen = 0;

    /* Store the simple data in the rr data */
    data->expire = 0;
    data->ttl = ttl;
    data->authoritative = authoritative;
    /* Get the rr type from the query string */
    if(datatype == MARA_DNS_NEG)
        rrtype = RR_SOA;
    else if(rtype == 255)
        rrtype = get_rtype(query);
    else 
        rrtype = rtype;
    if(rrtype == JS_ERROR) {
        js_dealloc(data);
	return JS_ERROR;
        } 
    if(rrtype < 0 || rrtype > 65535) {
        js_dealloc(data);
	return JS_ERROR;
        } 
    data->rr_type = rrtype;

    /* Create a js_string object to store the raw binary answer */
    if((new=js_create(value->unit_count + 1,value->unit_size)) == 0) {
        js_dealloc(data);
        return JS_ERROR;
	}
    if(js_copy(value,new) == JS_ERROR) {
        js_dealloc(data);
        js_destroy(new);
	return JS_ERROR;
	}
    /* And put a pointer to that js_string in the rr data */
    data->data = new;
    /* This is a new record, so the pointers to other records this uses
       are blank */
    data->ip = data->next = 0;

    /* Handle the case of there already being an element in the hash at the
       desired point */
    spot_data = mhash_get(hash,query);
    if(spot_data.value != 0) {
        /* If the action is to overwrite, delete the element that is
	   currently there */
        if(action == 1 || spot_data.datatype == MARA_DNS_NS) { /* Overwrite */
	    /* We also overwrite in the [should never happen] case of the
	       element in question being a MARA_DNS_NS element */
	    switch(spot_data.datatype) {
	        case MARA_DNSRR:
		case MARA_DNS_NEG:
		    mhash_undef(hash,query);
		    unlink_rr(spot_data.value,0);
		    break;
                case MARA_DNS_NS:
		    mhash_undef(hash,query);
		    unlink_closer(spot_data.value);
                    break;
		default: /* There should not be any other data types
		            in the hash */
		    return JS_ERROR;
                }
	    }
        else if(action == 2) { /* Append */
	    point = spot_data.value;
	    while(point->next != NULL)
	        point = point->next;
            point->next = data;
	    }
        }

    /* Store that data in the big hash, but only if it is appropriate
       to create a new element */
    if(action == 1 /* Overwrite */ || 
       spot_data.value == 0 /* Nothing to append to: Create new element */ ) {
        ret = mhash_put(hash,query,data,datatype);
	if(ret == JS_ERROR)
	    return JS_ERROR;
        }

    /* Now that we have copied the key to the big hash, add a pointer
       to that key to the data.  BTW, don't use mhash_get_immutable_key then
       change the value of the key string--dong so will mess things up
       badly (it will make it impossible to ever get that hash element
       again). */
   
    if(datatype == MARA_DNS_NEG || rtype != 255) {
	if((new_query = js_create(query->unit_count + 1,1)) == 0) {
            js_dealloc(data);
            js_destroy(new);
	    return JS_ERROR;
	    }
	if(js_copy(query,new_query) == JS_ERROR) {
	    js_destroy(new_query);
            js_dealloc(data);
            js_destroy(new);
	    return JS_ERROR;
	    }
	if(change_rtype(new_query,rrtype) == JS_ERROR) {
	    js_destroy(new_query);
            js_dealloc(data);
            js_destroy(new);
	    return JS_ERROR;
	    }
        data->query = new_query;
        } 
    else
        data->query = mhash_get_immutable_key(hash,query);

    if(ret == JS_ERROR) {
        js_dealloc(data);
        js_destroy(new);
	return JS_ERROR;
        } 
 
    return JS_SUCCESS;
    }

/* If we were given a negative answer, we have to add the fact that
   the record does not exist to our cache, and return the negative
   answer back to the stub resolver.
   Input: The query that they sent us;
          A pointer to the js_string with the uncompressed reply from the
          DNS server; offset from the beginning of that string where the
	  negative SOA reply begins; pointer to header of reply
   Output: JS_ERROR if something nasty happened, JS_SUCCESS if we managed
           to add the data to our cache.
   Global variables used: The dns cache, of course.
*/

int handle_negative_data(js_string *query, js_string *server_reply, 
	        int offset, q_header *header) {
    int query_type, len;
    uint32 ttl;
    int rdlength, ret;
    rr data;

    /* Determine the original query type they asked for */
    query_type = get_rtype(query);
    if(query_type == JS_ERROR)
	return JS_ERROR;

    /* Dealing with negativity is such a pain.
       What our code does is create a special element in the cache that
       can easily be used by the already existing udpnotfound routine.  
       We make the data an ordinary rr.  We can determine the expire by
       the ttd in the rr, and the contents the RR points to is the data
       that we show the end user.
     */

    /* Determine the length of the answer */
    len = dlabel_length(server_reply,offset);
    if(len == JS_ERROR)
	return JS_ERROR;
    /* Add the type to the length */
    len += 2;
    /* Go past the query, type, and class */
    offset += len + 2;
    /* Get the TTL */
    /* XXX: RFC1034 says you use the SOA minimum as the TTL.  Instead, we
       will use the TTL of the SOA record itself */
    ttl = js_readuint32(server_reply,offset);
    if(ttl == JS_ERROR)
	return JS_ERROR;
    offset += 4;
    /* Get the rdlength of the SOA record */
    rdlength = js_readuint16(server_reply,offset);
    if(rdlength == JS_ERROR)
	return JS_ERROR;
    /* Add the rddata to the big hash */
    offset += 2;
    ret = substring_add_rr(dnscache,server_reply,offset,rdlength,ttl,query,
	                    1,MARA_DNS_NEG);
    return ret;
    }

/* Given two pointers at differnet places in the same js_string object,
   both of which point to a dns-style domain name (in binary format),
   determine if the two objects point to the same domain name.
   Input: Pointer to the JS string we are doing the comparison of, index of
          the first point we start our comparison at, index of the second
	  point we start our comparison at.
   Output: JS_ERROR (-1) on fatal error, 0 if they do not match, 1 if
           they do match */
int cmp_dnames(js_string *js, int p1, int p2) {
    int temp, length;

    if(js_has_sanity(js) == JS_ERROR)
        return JS_ERROR;

    /* Make sure p1 is smaller than p2 */
    if(p1 > p2) {
        temp = p1;
	p1 = p2;
	p2 = temp;
	}
  
    /* If they point to the same place, they are the same (obviously) */
    if(p1 == p2)
        return 1;

    /* If p1 or p2 are out of bounds, then return error 
       Optimization: p2 is greater then p1, so we only have to check to see
       if p1 is less then zero */
    if(p1 < 0 || p1 > js->unit_count || p2 > js->unit_count)
        return JS_ERROR;

    length = dlabel_length(js,p1);

    /* They are not the same if they overlap */
    if(p1 + length > p2)
        return 0;

    /* They are not the same if they have different lengths */
    if(length != dlabel_length(js,p2))
        return 0;

    /* Make sure we are in bounds (optimization: we know p2 is greater than
       p1) */
    if(p2 + length > js->unit_count)
        return JS_ERROR;

    /* OK, now compare the two dlabels byte by byte */
    temp = 0;
    while(temp < length) {
        if(*(js->string + p1 + temp) != *(js->string + p2 + temp))
	    return 0;
        temp++;
        }
    
    /* There is no difference between the labels.  Return 1 */
    return 1;

    }

/* Query a nameserver for the answer to our question.  Add the answer
   (it being a closer nameserver or the question being asked)
   If answer found, add it to the cache.  If pointer to other name servers 
   given, add those pointers to this cache. 
   Input: IP of server we will contact, query to send to the server,
          bailiwick of the IP of the server we will contact
   Output: JS_ERROR on fatal error, -2 if the nameserver could not
           be contacted (timeout), -3 if the namserver refuses to give us the
	   the appropriate information, -4 if it is a lame delegation
	   (non-authoritative data), JS_SUCCESS if we find a name server,
	   2 if we got the final answer, 3 if we got the final answer and
	   it is a CNAME record
*/

int query_nameserver(int remote_ip, js_string *query, js_string *bailiwick) {
    /* Lots of cut and paste from askmara.c here */
    struct sockaddr_in dns_udp, server;
    int len_inet; /* Length */
    int s; /* Socket */
    int q; /* Used for binding, etc */
    int nspoint, offset, len;
    int type, class, rdlength;
    int cname_original_record = 0; /* Use a variable for two purposes:
				      - Determine if we cname'd this record
				      - If so, what was the Query ID of their
				        original query.
                                   */
    uint32 ttl;
    uint16 sid; /* Securly generated query id */
    uint16 ns_record_type; /* Whether this is a "host not found"/NXDOMAIN or a
			      NS referral */
    int record_added = 0; /* Boolean: has a record been added to the cache */
    js_string *outdata; /* Outgoing data */
    js_string *indata, *uindata; /* Incoming data (uncompressed version) */
    js_string *jsip; /* Place where we put the IP of the answer */
    uint32 ipq; /* IP which we pass the address of as an argument for 
		   determining the IP that a CNAME records points to */
    q_header header; /* header data */
    int counter,place,count,ret;
    /* Select() [which we are using *only* because of its timeout ability]
       requires four variables to be used */
    fd_set rx_set; int maxd; struct timeval tv; int n;
    /* We keep track of what NS servers are servers for a given
       branch of the "domain" tree in the answer */
    uint16 ns_record[32]; /* Who is going to have more than 32 NS records 
                             for a given domain name? */
    /* We keep track of that branch of the DNS "tree" the records are
       authoritative for in the answer */
    uint16 ns_domain[32];
    struct sockaddr dummy; /* "dummy" so we can call recurse_call when
			      trying to determine the IP for a CNAME 
			      record */
    mhash_e point; /* Used for the CNAME code */
    rr *spot;

    /* Initialize ns_record and ns_domain */
    ns_record[0] = 0;
    ns_domain[0] = 0;

    /* Allocate memory for some strings */
    if((indata = js_create(512,1)) == 0)
        return JS_ERROR;
    if((uindata = js_create(2048,1)) == 0) {
        js_destroy(indata);
        return JS_ERROR;
	}
    if((outdata = js_create(512,1)) == 0) {
        js_destroy(indata);
	js_destroy(uindata);
        return JS_ERROR;
	}

    /* Create a UDP client socket */
    if((s = socket(AF_INET,SOCK_DGRAM,0)) == -1)
        goto cleanup;

    /* Create a server socket address to use with sendto() */
    memset(&server,0,sizeof(server));
    server.sin_family = AF_INET;
    server.sin_port = htons(53); 
    if((server.sin_addr.s_addr = htonl(remote_ip)) == INADDR_NONE)
        return JS_ERROR;
    /* Create a secure psudo-random port number to bind to.
       Give credit where credit is due: This is loosely modelled after
       an example UDP client I found on http://www.pont.net */
    memset(&dns_udp,0,sizeof(dns_udp));
    dns_udp.sin_family = AF_INET;
    /* XXX read some kvars which determine the range of this,
           also, find out how to change the local port number */
    dns_udp.sin_port = htons(15000 + (srng() & 4095)); 
    /* XXX also make this user-configurable */
    dns_udp.sin_addr.s_addr = htons(INADDR_ANY);

    len_inet = sizeof(dns_udp);

    /* Format a DNS request */
    sid = srng();
    header.id = sid;
    header.qr = 0;
    header.opcode = 0;
    header.aa = 0;
    header.tc = 0;
    header.rd = 0; /* We do not want recursion when sending out a query 
		      from the recursor */
    header.ra = 0;
    header.z = 0;
    header.rcode = 0;
    header.qdcount = 1;
    header.ancount = 0;
    header.nscount = 0;
    header.arcount = 0;

    /* Make a beginning of a DNS query from that header */
    if(make_hdr(&header,outdata) == JS_ERROR)
        goto cleanup;
    if(js_append(query,outdata) == JS_ERROR)
        goto cleanup;
    if(js_adduint16(outdata,1) == JS_ERROR) /* Adding class to query */
        goto cleanup;

    /* OK, the DNS request has been formed.  Send it out */
    /* Bind to a secure psudo-random address and port */
    if(bind(s,(struct sockaddr *)&dns_udp,sizeof(dns_udp)) < 0)
	goto cleanup;
    /* And send, on the same socket, the message to the server */
    if(sendto(s,outdata->string,outdata->unit_count,0,
       (struct sockaddr *)&server,len_inet) < 0)
        goto cleanup;

    /* Wait for a reply from the DNS server */
    FD_ZERO(&rx_set);
    FD_SET(s,&rx_set);
    maxd = s + 1;
    tv.tv_sec = 2;
    tv.tv_usec = 0;
    n = select(maxd,&rx_set,NULL,NULL,&tv);
    if(n == -1)  /* select error */ {
	close(s);
        goto cleanup;
        }
    if(n == 0) /* Timeout */ {
	close(s);
        goto minus2;
        }
    /* Get the actual reply from the DNS server */
    if((count = recvfrom(s,indata->string,indata->max_count,0,
                         (struct sockaddr *)&dns_udp,&len_inet)) < 0) {
	close(s);
        goto minus3;
        }
    /* Now that we are done with the socket, close it */
    close(s);
    /* Process the reply */
    indata->unit_count = count;
    decompress_data(indata,uindata);
    if(read_hdr(uindata,&header) == JS_ERROR) 
        goto minus3;
 
    /* OK, there are two school of thoughts here.  If we get a different
       Query ID than the one we asked for, there are two possibilities here:
 
       * Somone who should not has sent us a packet in an attampt to spoof
         us a packet.
 
       * For whatever reason, this particular recvfrom() got the wrong
         packet.
 
       In terms of security against a spoofer, we have two choices:
 
       * We allow a spoofer to perform a DOS attack against us.
 
       * We allow a spoofer to have multiple chances to sucessfully
         spoof a packet sent to us.
 
       I opt to allow a spoofer to DOS us, since coding this is easier.
 
     */
 
    if(header.id != sid)
        return JS_ERROR;

    /* Let us find the offset of the first answer */
    count = header.qdcount; 
    offset = 12; /* Point it at the first question/answer */
    while(count > 0) {
        len = dlabel_length(uindata,offset);
        if(len == JS_ERROR)
            goto cleanup;
        offset += len;
        offset += 4; /* Move past the query type and query class */
	count--;
	}

    /* Determine if this is an answer or a referral */
    if(header.ancount > 0) {
        /* If this is an answer, add the answer to our cache */
        /* We get the following pieces of information:
           - The Time to live (and its delta from right now)
	   - The exact query this is an answer for
	   - A js string object which is the answer for the query 
	     (newly created)
        */
	nspoint = 1; /* We use ns point here to tell us if we are pointing
	                to the first record or to a later record */
        counter = header.ancount;
        do {
            /* Determine the length of the answer */
	    len = dlabel_length(uindata,offset);
	    if(len == JS_ERROR)
	        goto cleanup;
	    /* Add the type to the length */
	    len += 2;
	    /* Make sure that the answer is the same as the query.
	       This routine also folds case as appropriate */
	    if(substring_issame_case(uindata,offset,len,query) != 1) {
		/* Perhaps this is a CNAME answer, which we also add
		   to the cache as appropriate */
		cname_original_record = get_rtype(query);
		if(cname_original_record == JS_ERROR)
		    goto cleanup;
		if(change_rtype(query,RR_CNAME) == JS_ERROR)
		    goto cleanup;
		if(substring_issame_case(uindata,offset,len,query) != 1)
	            goto minus2;
		/* If it is a CNAME answer, DNS servers have this way
		   of politely also giving us the IP the CNAME points to
		   in the same query.  Because of how MaraDNS is structured,
		   this information does not help us, and is, in fact, an
		   error condition if we try grokking the helpful additional
		   answer.  Hence, we don't look at records after a CNAME
		   record in the answer */
		counter = 0;
	        }
            /* Go past 'query', 'type', and 'class' */
	    offset += len + 2;
            /* Get the ttl */
	    ttl = js_readuint32(uindata,offset);
	    
	    if(ttl == JS_ERROR)
	        goto minus2;
            offset += 4;
            /* Get the rdlength */
	    rdlength = js_readuint16(uindata,offset);
	    if(rdlength == JS_ERROR)
	        goto minus2;
            /* Add the rddata to the big hash */
	    offset += 2;
            ret = substring_add_rr(dnscache,uindata,offset,rdlength,ttl,
	                           query,nspoint,MARA_DNSRR);
	    /* If the record in question is a CNAME, and we originally wanted
	       an A record, add (one of) the A record(s) that corresponds to
	       the CNAME record in question */
	    if(cname_original_record == RR_A) {
		/* Determine the IP that the CNAME record points to */
		if((jsip = js_create(256,1)) == 0)
		    goto cleanup;
		ipq = 0;
		/* Find out what the CNAME answer is so we can query for it */
	        point = mhash_get(dnscache,query);	
		if(point.value != 0 && point.datatype == MARA_DNSRR) {
		    spot = point.value;
		    if(spot->data != 0 && spot->rr_type == RR_CNAME) {
			if(js_copy(spot->data,jsip) == JS_ERROR)
			    goto cleanup;
			if(js_adduint16(jsip,RR_A) == JS_ERROR)
			    goto cleanup;
		        recurse_call(0,0,dummy,jsip,0,0,&ipq);
		        }
		    }
		/* Make the data pointed to by ipq a jsip 
		   (this is a DAV) */
		if(ipq != 0) {
		    *(jsip->string) = (ipq & 0xff000000) >> 24;
		    *(jsip->string + 1) = (ipq & 0x00ff0000) >> 16;
		    *(jsip->string + 2) = (ipq & 0x0000ff00) >> 8;
		    *(jsip->string + 3) = (ipq & 0x000000ff);
                    jsip->unit_count = 4;
                    mhash_add_ip(dnscache,query,jsip);
		    }
	        }
            nspoint = 2;
	    counter--;
	    offset += rdlength;
	    if(cname_original_record != 0) {
	        if(change_rtype(query,cname_original_record) == JS_ERROR)
		    goto cleanup;
	        }
	    } while(counter > 0);

	js_destroy(indata);
	js_destroy(uindata);
	js_destroy(outdata);
	if(cname_original_record == 0) /* If this was not a CNAME answer */
	    return 2; /* Final non-CNAME answer given */
	else
	    return 3; /* CNAME answer given */
	}
   
    /* It is a referral --or-- it is a "host not here" */
    
    /* If there are no authority records either, then it is an error */
    if(header.nscount == 0) 
        goto minus2;

    /* Offset now points at the first authority (NS) record */
    count = header.nscount;
    nspoint = 0;
    /* We only look at the first 29 NS records */
    if(count > 29)
       count = 29;

    /* Start filling up the ns_record array with pointers to the
       various authority records that are ns_records.  Later on,
       we will see which ns records have corresponding records in
       the additional records section. */
       
    nspoint = 0;
    while(nspoint < count) {
        /* Determine the length of the answer returned */
        len = dlabel_length(uindata,offset);
	if(len == JS_ERROR)
	    goto cleanup;
        /* Point to the question they asked 
	   (but only if this is a NS record) */
        /* Handle SOA records (read: negative caching) in the NS section */
	ns_record_type = js_readuint16(uindata,offset + len);
	/* If this is a "that does not exist" answer, we give the answer
	   special processing. */
	if(nspoint == 0 && ns_record_type == RR_SOA) { 
	    return handle_negative_data(query,uindata,offset,&header);
	    }
	if(ns_record_type == RR_NS && offset > 0)
            ns_domain[nspoint] = offset;
        else {
	    ns_domain[nspoint] = 1; /* Invalid data */
            }
        /* Answers that are out of bailiwick are invalid */
	if(offset_bailiwick(uindata,offset,bailiwick) != 1) {
	    ns_domain[nspoint] = 1; /* Invalid data */
            }
        ns_domain[nspoint + 1] = 0; /* End of ns records */
        offset += len + 8; /* Go past name, type, class, and ttl */
	len = js_readuint16(uindata,offset); /* rdlength */
	if(len == JS_ERROR)
	    goto cleanup;
        /* We also keep a note of what the NS record is for
	   this domain */
        if(ns_domain[nspoint] > 2) { /* If this is a valid NS record */
            ns_record[nspoint] = offset + 2;
	    }
        else {
	    ns_record[nspoint] = 1; /* invalid data */
	    }
	ns_record[nspoint + 1] = 0;
	offset += len + 2; /* Go past rdlength and rddata */
        nspoint++; 
        }

    /* This string is used to pass arguments to add_closer_jsip */
    if((jsip = js_create(6,1)) == 0)
        goto cleanup;

    /* Start looking at the additional records.  See which ones are IPs
       which match ns records.  When we see such an IP, we add the record
       to the the big cache. */

    count = header.arcount;
    while(count > 0) {
        /* Determine the length of the question we are answering */
	len = dlabel_length(uindata,offset);
	if(len == JS_ERROR)
	    goto cleanup_jsip;
        /* We will only add this record to the cache if:
	   1) The answer is a NS record 
	   2) The answer is in bailiwick (check done above)
	   3) There is a ns record above corresponding to the answer
         */
	type = js_readuint16(uindata,offset + len);
	class = js_readuint16(uindata,offset + len + 2);
	ttl = js_readuint32(uindata,offset + len + 4);
	rdlength = js_readuint16(uindata,offset + len + 8);
	if(type == JS_ERROR || class == JS_ERROR || ttl == JS_ERROR || 
	   rdlength == JS_ERROR)
	    goto cleanup_jsip;
        /* Move the offset pointer to point to the beginning of the
	   rddata */
        if(type != RR_A || class != 1 || rdlength != 4) {
	    offset += len + rdlength + 10; /* Rdlength plus ten byte header,
	                                      plus length of domain name
					      label */
            count--;
	    continue;
	    }
        /* Start comparing the dlabel for this RR with the various NS
	   RRs in the authority section */
        nspoint = 0;
	while(ns_record[nspoint] != 0) {
	    /* Give the jsip string a value */
	    if(js_substr(uindata,jsip,len + offset + 10,4) == JS_ERROR)
	        goto cleanup_jsip;
            /* If the domain name label for this node share the same
	       name as a ns record */
	    if(ns_record[nspoint] > 8 && 
	        cmp_dnames(uindata,ns_record[nspoint],offset) == 1) {
	        if(record_added == 0) {
	            record_added = 1;
		    /* Handle case */
		    check_case_of_answer(uindata,query,ns_domain[nspoint]);
		    if(add_closer_jsip_offset(uindata,ns_domain[nspoint],
		                              jsip,OVERWRITE) == JS_ERROR)
		        goto cleanup_jsip;
		    }
                else {
		    /* Handle case */
		    check_case_of_answer(uindata,query,ns_domain[nspoint]);
		    if(add_closer_jsip_offset(uindata,ns_domain[nspoint],
		                              jsip,APPEND) == JS_ERROR)
		        goto cleanup_jsip;
                    }
                /* Mark this as "already added" */
		ns_domain[nspoint] = 2;
		ns_record[nspoint] = 2;
                }
            nspoint++;
	    }
        offset += len + rdlength + 10;
	count--;
        }
 
    /* All of the glued records have been added.  Now add the glueless 
       records to the cache */

    nspoint = 0;
    while(ns_record[nspoint] != 0) {
        /* If the record is valid and has not been added */
        if(ns_record[nspoint] > 8) {
	    if(record_added == 0) {
	        record_added = 1;
		/* Handle case */
		check_case_of_answer(uindata,query,ns_domain[nspoint]);
	        if(add_closer_js_offset(uindata,ns_domain[nspoint],
				ns_record[nspoint],OVERWRITE) == JS_ERROR)
		    goto cleanup_jsip;
		}
            else {
		/* Handle case */
		check_case_of_answer(uindata,query,ns_domain[nspoint]);
	        if(add_closer_js_offset(uindata,ns_domain[nspoint],
	                              ns_record[nspoint],APPEND) == JS_ERROR)
		    goto cleanup_jsip;
                }
            }
        nspoint++;
        } 

    js_destroy(jsip);
    js_destroy(indata);
    js_destroy(uindata);
    js_destroy(outdata);
    return JS_SUCCESS;

    cleanup_jsip:
        js_destroy(jsip);
    cleanup:
        js_destroy(indata);
	js_destroy(uindata);
	js_destroy(outdata);
	return JS_ERROR;

    minus2_jsip:
        js_destroy(jsip);
    minus2:
        js_destroy(indata);
	js_destroy(uindata);
	js_destroy(outdata);
	return -2;

    minus3_jsip:
        js_destroy(jsip);
    minus3:
        js_destroy(indata);
	js_destroy(uindata);
	js_destroy(outdata);
	return -3;

    }

/* Give answer: Give an answer from the cache.
   Input: the pointer to the element in the hash with the data, the data type
          of the element in the hash with the data, a pointer to where
	  we store the IP we get (if appropriate), the query the end user 
	  sent to the program, a pointer to wher ein the hash we found the
          answer, the id of the query the end-user sent, the fd of the socket
	  the query is on, the ip and other related informaton about the 
	  client who sent the query to us, the recursion depth we are 
	  currently at, the glueless level we are currently at
   Output: 0 if we didn't find any data, JS_SUCCESS if we successfully 
           sent data to the end user -or- we set ipret
*/

int give_answer(void *value, int datatype, uint32 *ipret, js_string *query,
                js_string *hash_pointer,
                int id, int sock, struct sockaddr client, int queries_sent,
		int glueless_level, int cname_answer) {


    rr *lookatrr;
#ifdef DEBUG
    printf("Entering give_answer...\n");
#endif

        /* Make sure the RR has not expired */
	lookatrr = value;
	/* If the data expires (is not 0), and the data has expired... */
	if(lookatrr->expire != 0 && lookatrr->expire < time(0)) {
	    /* Then destroy the data */
#ifdef DEBUG
	    printf("RR has expired from cache, unlinking\n");
#endif
	    /* Make the read-only lock a read-write lock */
	    read2write_lock();
	    /* Delete the answer from the cache */
	    lookatrr = mhash_undef(dnscache,hash_pointer);
	    unlink_rr(lookatrr,0);
	    write_unlock();
	    /* And re-run recurse_call with the revised cache */
	    recurse_call(id,sock,client,query,queries_sent + 1,
		         glueless_level,ipret);
	    goto success;
	    }

	/* If we are set up to return an answer to the client instead
	   of placing the answer in a pointer (used for out-of-bailiwick 
	   and cname-to-ip lookups), do so */
        if(ipret == NULL) {
	    read2write_lock();
	    if(datatype == MARA_DNS_NEG)
	        udpnotfound(lookatrr,id,sock,&client,query,0);
            else
                udpsuccess(value,id,sock,&client,query);
	    write2read_lock();
	    }
	else {
	    lookatrr = value;
	    /* Make sure this is an A record */
	    if(lookatrr->rr_type != RR_A && cname_answer != 1) {
		read_unlock();
		*ipret = 0;
		goto success;
	        }
	    /* If it is a cname answer, guess what: We get to find out 
	       what IP that cname (eventually) points to */
	    if(cname_answer == 1) {
		/* Copy over lookatrr->data */
		if(js_copy(lookatrr->data,hash_pointer) == JS_ERROR) 
		    goto cleanup;
                if(js_adduint16(hash_pointer,RR_A) == JS_ERROR)
                    goto cleanup;
                read_unlock(); 
                recurse_call(id,sock,client,hash_pointer,queries_sent + 1,
			glueless_level + 1,ipret);
		goto success;
	        }
	    /* Convert spot_data.value->data in to a uint32 */
#ifdef DEBUG
	    printf(" Returning IP: ");
	    show_esc_stdout(lookatrr->data);
	    printf("\n");
#endif
	    *ipret = ((*(lookatrr->data->string) << 24) & 0xff000000) |
	             ((*(lookatrr->data->string + 1) << 16) & 0x00ff0000) |
	             ((*(lookatrr->data->string + 2) <<  8) & 0x0000ff00) |
	             (*(lookatrr->data->string + 3) & 0x000000ff);
            }

    read_unlock();

    success:
        return JS_SUCCESS;
    cleanup:
        return JS_ERROR;
    }

/* Recurse thread: This is a special routine that launches its own thread 
   input: Pointer to dnsreq query
   output: JS_SUCCESS (never looked at--this is a detached thread)
*/

int recurse_thread(dnsreq *req) {

    int ret;

    ret = recurse_call(req->id, req->sock, req->client, req->query, 0, 0, 
	               NULL);
    js_destroy(req->query);
    js_dealloc(req);
    return ret;

    }

/* We made this a separate function so that we can call it recursivly 
   without having to allocate memory on the heap for req objects.
   Input: id of request, socket of request, location request came from,
          desired DNS query, recursion_depth (which we use to limit the
	  number of queries), gluless_level (the number of times we had
	  to perform a glueless lookup), ipret: If this is a request for
	  a glueless IP, return the IP here.  If this is NULL, give the
	  client the answer to the question they asked.
   Output: JS_SUCCESS, I believe
*/

int recurse_call(int id, int sock, struct sockaddr client, 
                 js_string *query, int queries_sent, int glueless_level,
		 uint32 *ipret) {

    mhash_e spot_data;
    js_string *copy;
    closer *cpoint;
    uint32 nsip;
    int result, qtype, case_folded = 0, cname_answer = 0;
    /* Variables used to copy over the chain of closer name servers */
    closer *local_c, *local_c_head, *local_c_save;
    uint32 *i32_copy;
    js_string *jstr_copy, *glueless_query, *lower;
    rr *lookatrr;

#ifdef DEBUG 
    printf("D: %d G: %d I: %p Requested query: ",queries_sent,glueless_level,ipret);
    show_esc_stdout(query);
    printf("\n");
#endif

    /* Make sure we haven't overloaded the recursion depth or maximum glueless 
       level */
    if(queries_sent > 32 || glueless_level > 4)
        return JS_ERROR;

    if(ipret != NULL)
	*ipret = 0;

    /* See if we have to do case-insensitive searching */
    if((lower = js_create(query->unit_count + 3,1)) == 0)
        return JS_ERROR;
    if(js_copy(query,lower) == JS_ERROR)
	goto cleanup_nojs;
    /* leave on error */
    if(case_folded == JS_ERROR) 
	goto cleanup_nojs;
    
    /* We deal with the case of RR_ANY thusly:
       1) Look for RR_A and RR_MX for the query in question
       2) Use the udpany call to "sew" together the data
       3) Return a SOA if nothing was found */
    qtype = get_rtype(query);
    if(qtype == RR_ANY) {
        /* Make sure the RR_A and RR_NS are placed in the dns cache */
        change_rtype(query,RR_A);
        recurse_call(id,sock,client,query,queries_sent + 1,
	             glueless_level,&nsip);
        change_rtype(query,RR_MX);
        recurse_call(id,sock,client,query,queries_sent + 1,
	             glueless_level,&nsip);
        change_rtype(query,RR_ANY);
	/* If something was found, we have success */
        if(udpany(id,sock,&client,query,3,dnscache) == JS_SUCCESS) {
            js_destroy(lower);
	    return JS_SUCCESS;
	    }
        /* Return a SOA answer if an A or MX answer was not found. */
	lookatrr = 0;
        change_rtype(query,RR_A);
	spot_data = mhash_get(dnscache,query);
	if(spot_data.value != 0 && spot_data.datatype == MARA_DNS_NEG)
	    lookatrr = spot_data.value;
        else {
	    change_rtype(query,RR_MX);
	    spot_data = mhash_get(dnscache,query);
	    if(spot_data.value != 0 && spot_data.datatype == MARA_DNS_NEG)
	        lookatrr = spot_data.value;
            }
        if(lookatrr != 0)
	    udpnotfound(lookatrr,id,sock,&client,query,255);
        js_destroy(lower);
	return JS_SUCCESS;
        }

    /* Allocate memoey for the 'copy' string */
    if((copy = js_create(257,1)) == 0) 
        goto cleanup_nojs;

    /* Look for query in cache */
    read_lock();
    spot_data = mhash_get(dnscache,lower); 
   
    /* If there is a "this host does not exist" in the cache, return that
       answer */
    if(spot_data.value != 0 && spot_data.datatype == MARA_DNS_NEG) {
        if(give_answer(spot_data.value,spot_data.datatype,ipret,query,
	               lower,id,sock,client,queries_sent,
		       glueless_level,0) == JS_SUCCESS)
            goto success;
        }

    /* Return cached data if found */
    if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
        if(give_answer(spot_data.value,spot_data.datatype,ipret,query,
	               lower,id,sock,client,queries_sent,
		       glueless_level,0) == JS_SUCCESS)
            goto success;
	}

    /* If it is not found, see if we have a CNAME record instead */
    if(js_copy(lower,copy) == JS_ERROR)
	goto cleanup;
    if(change_rtype(copy,RR_CNAME) == JS_ERROR)
        goto cleanup;
    spot_data = mhash_get(dnscache,copy);
    if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
        if(give_answer(spot_data.value,spot_data.datatype,ipret,query,
	               copy,id,sock,client,queries_sent,
		       glueless_level,1) == JS_SUCCESS)
            goto success;
	}

    read_unlock();

    /* If we still haven't found something, perhaps a lower-case version of
       the same exist */
    case_folded = fold_case(lower);
    /* If we could fold the case, look for a lower-case version of the same */
    if(case_folded == 1) {
        spot_data = mhash_get(dnscache,lower); 
   
        /* If there is a "this host does not exist" in the cache, return that
           answer */
        if(spot_data.value != 0 && spot_data.datatype == MARA_DNS_NEG) {
            if(give_answer(spot_data.value,spot_data.datatype,ipret,query,
	               lower,id,sock,client,queries_sent,
		       glueless_level,0) == JS_SUCCESS)
                goto success;
            }

        /* Return cached data if found */
        if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
            if(give_answer(spot_data.value,spot_data.datatype,ipret,query,
	               lower,id,sock,client,queries_sent,
		       glueless_level,0) == JS_SUCCESS)
                goto success;
	    }
        }

#ifdef DEBUG
    printf("Not found in cache, bugging nameserver.....\n");
#endif

    /* OK, we don't have data in the cache.  See which nameserver we
       have to bug */
    if(js_copy(query,copy) == JS_ERROR)
        goto cleanup;

    /* See if there is a name server in the cache for the exact query we are
       asking for */
    if(change_rtype(copy,252) == JS_ERROR)
        goto cleanup;
    /* OK, look for the "lowest" branch in the tree which will lead us to 
       the data we are seeking */
    do {
        read_lock();
	spot_data = mhash_get(dnscache,copy);
	/* Query nameserver if found */
	if(spot_data.value != 0 && spot_data.datatype == MARA_DNS_NS) {
	    cpoint = spot_data.value;
	    /* If the data expires (is not 0), and the data has expired... */
	    if(cpoint->ttd != 0 && cpoint->ttd < time(0)) {
		/* The destroy the data */
		/* Make the read-only lock a read-write block */
#ifdef DEBUG
	        printf("Name Server has expired from cache, unlinking\n");
#endif
		read2write_lock();
		cpoint = mhash_undef(dnscache,copy);
		unlink_closer(cpoint);
		write_unlock();
		/* Re-run this routine with the revised cache */
		recurse_call(id,sock,client,query,queries_sent + 1,
			     glueless_level,ipret);
		goto success;
	        }

	    /* Code that makes a local copy of the entire 
	       cpoint chain, so that other threads can use (or destroy) the
	       original */
	    if((local_c = js_alloc(1,sizeof(closer))) == 0)
	        goto cleanup;
            local_c_head = local_c;
	    while(local_c != NULL) {
                local_c->num_elements = cpoint->num_elements;
	        local_c->ttd = cpoint->ttd;
	        switch(cpoint->datatype) {
	            case RR_A:
                        if((i32_copy = js_alloc(1,sizeof(uint32))) == 0)
		            goto cleanup;
                        *i32_copy = *(uint32 *)(cpoint->data);
		        local_c->datatype = RR_A;
                        local_c->data = i32_copy;
		        break;
		    case RR_NS:
		        if(js_has_sanity(cpoint->data) == JS_ERROR)
		            goto cleanup;
                        jstr_copy = cpoint->data;
                        jstr_copy = js_create(jstr_copy->unit_count + 1,1);
		        if(js_copy == 0)
		            goto cleanup;
                        if(js_copy(cpoint->data,jstr_copy) == JS_ERROR) {
		            js_destroy(jstr_copy);
			    goto cleanup;
			    }
                        local_c->datatype = RR_NS;
		        local_c->data = jstr_copy;
		        break;
                    default:
		        local_c->datatype = 10; /* NULL RR number, since this
		                                   data is invalid */
                        local_c->data = NULL;
                    }
		if(cpoint->next != NULL) {
                    local_c_save = local_c;
	            if((local_c = js_alloc(1,sizeof(closer))) == 0)
	                goto cleanup;
		    local_c_save->next = local_c;
                    cpoint = cpoint->next;
		    }
                else {
		    local_c->next = NULL;
		    break;
		    }
	        }

            /* We have copied the data, so we can now unlock the global
	       data */
            read_unlock();
	    local_c = local_c_head;
	    /* Select a NS server at random (weak RNG because this is just
	       to be fair to the name servers) */
	    result = random() % local_c->num_elements; /* I know, using
							  an unreleared
							  variable as a
							  temporary counter
		                                       */
#ifdef DEBUG
            printf("Going to: %d (out of %d)\n",result,local_c->num_elements);
#endif
            while(result > 0) {
                local_c = local_c->next;
		if(local_c == NULL)
		    local_c = local_c_head;
                result--;
                }
#ifdef DEBUG
		show_esc_stdout(copy);
		printf("\n");
#endif
            local_c_save = local_c;
	    result = 0;
	    /* Go through all of the NS servers we have an IP for.  Since
	       we are using a circular linked list, this code is a little
	       tricky. */
	    do {
                if(local_c->datatype == RR_A) {
                    nsip = *(uint32 *)(local_c->data);
                    /* Query the nameserver for the answer */
#ifdef DEBUG
                    printf("Querying nameserver ");
                    debug_show_ip(nsip);
	            printf("\n");
		    fflush(stdout);
#endif /* DEBUG */
	            result = query_nameserver(nsip,query,copy);
#ifdef DEBUG
	            printf("Result of query: %d\n",result);
#endif /* DEBUG */
                    }
		/* Go to the next nameserver on the list */
		local_c = local_c->next;
                if(local_c == NULL) 
                    local_c = local_c_head;
                } while (result <= 0 && local_c != local_c_save);
            /* If we did not find any glued name servers, do a DNS lookup 
	       for the glueless server(s) */
	    if(result <= 0) { /* If nothing found yet */
		/* Go to randomly selected element above */
		local_c = local_c_save;
		/* Do an A lookup for the glueless nameserver in question,
		   then, if successful, query that nameserver */
		do {
	            if(local_c->datatype == RR_NS) {
                        /* Look for the ip for the name */
		        /* Create a copy of the destired query */
		        glueless_query = local_c->data;
		        if(js_has_sanity(glueless_query) == JS_ERROR)
			    goto cleanup;
		        if((glueless_query = 
			    js_create(glueless_query->unit_count + 2,1)) == 0)
                            goto cleanup;
		        if(js_copy(local_c->data,glueless_query) == JS_ERROR) {
			    js_destroy(glueless_query);
			    goto cleanup;
		            }
		        /* Try to get the number */
                        if(recurse_call(id,sock,client,glueless_query,
			   queries_sent + 1, glueless_level + 1,&nsip) ==
                           JS_ERROR) {
			    js_destroy(glueless_query);
			    goto cleanup;
		            }
		        /* If we got an IP, query that nameserver, update the 
		           cache */
		        if(nsip != 0)
	                    result = query_nameserver(nsip,query,copy);
                        }
		    /* Go to the next nameserver on the list */
		    local_c = local_c->next;
                    if(local_c == NULL) 
                        local_c = local_c_head;
                    } while(result <= 0 && local_c != local_c_save);
                }
	    js_destroy(copy);
	    js_destroy(lower);
	    /* Since we are done with out local copy of the data, we can
	       get rid of it */
            unlink_closer(local_c_head);
	    if(result > 0) /* If a nameserver was found,
	                      then do a lookup again based 
			      on the revised cache */
	        recurse_call(id,sock,client,query,queries_sent + 1,
	                     glueless_level,ipret);
	    return JS_SUCCESS;
	    }
        read_unlock();
        } while(bobbit_label(copy) > 0);

    /* We could not find a "closer" name server, which is an error
       condition */
    /* Clean up the request */
    success:
        js_destroy(copy);
        js_destroy(lower);
        return JS_SUCCESS;

    cleanup:
        js_destroy(copy);
    cleanup_nojs:
	js_destroy(lower);
	return JS_ERROR;
    }

/* Routine which prepares the launch of a detached thread which will 
   recursivly look for the DNS name in question 
   Input:  id of query, socket of query, sockaddr struct of query,
           js_string with host name they are looking for
   Output: JS_ERROR on error, JS_SUCCESS on success
*/

int launch_thread(int id, int sock, 
	struct sockaddr_in client, js_string *query) {
    /* We have to copy the js_string over because the thread needs her
       own copy */
    js_string *copy;
    dnsreq *req;
    pthread_t thread;
    pthread_attr_t attr;

    /* Set up the attributes for the thread.  Make this a detached 
       thread because we don't look at the return value (the thread
       is talking to the client and maybe updating the cache) */
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);

    /* We need to create some thread-safe data structures.  
       We have a special dnsreq data structure that we use because a 
       thread, when initialized, can only be initialized with a single 
       argument. */

    /* Create the req data structure */
    if((req = js_alloc(1,sizeof(dnsreq))) == 0)
        return JS_ERROR;

    /* Create the 'copy' js_string */
    if((copy = js_create(query->unit_count + 3,1)) == 0) {
        js_dealloc(req);
	return JS_ERROR;
	}

    /* Copy over query to 'copy' */
    if(js_copy(query,copy) == JS_ERROR) {
        js_dealloc(req); js_destroy(copy);
	return JS_ERROR;
	}

    /* Fill out the rest of the req structure */
    req->id = id;
    req->sock = sock;
    req->client = client;
    req->query = copy;

    /* Spawn the thread */
#ifdef NOTHREAD
    recurse_thread(req);
#else
    if(pthread_create(&thread,&attr,(void *)recurse_thread,(void *)req) != 0)
        {
        js_dealloc(req); js_destroy(copy); return JS_ERROR;
	}
#endif /* NOTHREAD */

    /* Leave...it is now up to the thread to handle the query and free
       memory used */
    return JS_SUCCESS;

    }

/* Add an IP as a server "closer" to the answer for a given
   branch of the DNS tree 
   Input:  js_string of zone (raw UDP format with 2-byte suffix), 
           js_string with ddip, 
	   action to take if a node is already here:
	   1) Overwrite 2) Append 3) "Mask" currect table
   Output: JS_ERROR on fatal error, JS_SUCCESS on success
*/

int add_closer_jsddip(js_string *zone, js_string *ddip, int if_exists) {
    closer *close;
    js_string *ipjs;
    uint32 ip, *ipp;
    closer *point, *point_save;
    mhash_e spot_data;

    /* Allocate the memory for a new spot on the dnscache hash */
    if((close = js_alloc(1,sizeof(closer))) == 0)
        return JS_ERROR;

    /* Allocate the memory for the IP we will put on the hash */
    if((ipjs = js_create(6,1)) == 0) {
        js_dealloc(close);
	return JS_ERROR;
	}

    /* ddip_2_ip uses js_atoi, which means we have to set the encoding */
    ddip->encoding = MARA_LOCALE;

    /* Convert their ddip in to an ip */
    if(ddip_2_ip(ddip,ipjs,0) == JS_ERROR)
        goto cleanup;

    /* Make sure the raw zone is of the right rtype */
    if(change_rtype(zone,252) == JS_ERROR)
        goto cleanup;

    /* Convert ipjs in to a format that sendto() can use */
    if(ipjs->unit_count < 4)
        goto cleanup;
    ip = (*(ipjs->string) << 24) | (*(ipjs->string + 1) << 16) | 
         (*(ipjs->string + 2) << 8) | *(ipjs->string + 3);

    /* Fill up the "close" structure */
    close->num_elements = 1;
    close->datatype = RR_A;
    if((ipp = js_alloc(1,sizeof(uint32))) == 0)
        goto cleanup;
    *ipp = ip;
    close->data = ipp;
    close->next = NULL;
    close->ttd = time(0) + DEFAULT_TTL; /* XXX: Use the server-supplied ttl */
    close->ttd = 0; /* Hack: The root name server IPs never expire */

    /* Add the data to the hash */
    write_lock();
    /* See if the data is already there.  If so, act based on the value
       of if_exists */
    spot_data = mhash_get(dnscache,zone);
    if(spot_data.value != 0 && spot_data.datatype == MARA_DNS_NS) {
        point = spot_data.value;
	switch(if_exists) {
	    case 1: /* Overwrite */
	        /* Destroy the pointer to the element in the hash */
		mhash_undef(dnscache,zone);
		/* Destroy the element itself.  Go down the
		   linked list ans destroy each element one by one,
		   starting at the top. */
	        while(point != 0) {
	           point_save = point->next;
		   if(point->datatype == RR_A && point->data != 0) 
		       js_dealloc(point->data);
                   else if(point->datatype == RR_NS && point->data != 0)
		       js_destroy(point->data);
                   else if(point->data != 0)
		       js_dealloc(point->data);
                   js_dealloc(point);
		   point = point_save;
                   }
                break; 
	    case 2: /* Append */
                while(point->next != 0) 
	            point = point->next; 
                point->next = close;
		return JS_SUCCESS;
            /* XXX to do: mask */
	    }
        }
    /* This error condition should never happen */
    else if(spot_data.value != 0)
        return JS_ERROR;
    /* It's a (possibly newly) empty spot--add a new entry to the hash 
       (w/ error checking) */
    else if(mhash_put(dnscache,zone,close,MARA_DNS_NS) == JS_ERROR) {
        write_unlock();
	goto cleanup;
	}
    /* Make sure we do not overflow the cache */
    custodian();
    write_unlock();

    return JS_SUCCESS;
    cleanup_ipp:
        js_dealloc(ipp);
    cleanup:
        js_destroy(ipjs);
	js_dealloc(close);
	return JS_ERROR;

    }

/* Add an IP as a server "closer" to the answer for a given
   branch of the DNS tree 
   Input:  js_string of zone (raw UDP format with 2-byte suffix), 
           js_string with ddip, 
	   action to take if a node is already here:
	   1) Overwrite 2) Append 3) "Mask" currect table
   Output: JS_ERROR on fatal error, JS_SUCCESS on success
*/

int add_closer_ipv4pair(js_string *zone, ipv4pair *pair, int if_exists) {
    closer *close;
    uint32 ip, *ipp;
    closer *point, *point_save;
    mhash_e spot_data;

    /* Allocate the memory for a new spot on the dnscache hash */
    if((close = js_alloc(1,sizeof(closer))) == 0)
        return JS_ERROR;

    /* Make sure the raw zone is of the right rtype */
    if(change_rtype(zone,252) == JS_ERROR)
        goto cleanup;

    /* Convert the data (real simple in this case) */
    ip = pair->ip;

    /* Fill up the "close" structure */
    close->num_elements = 1;
    close->datatype = RR_A;
    if((ipp = js_alloc(1,sizeof(uint32))) == 0)
        goto cleanup;
    *ipp = ip;
    close->data = ipp;
    close->next = NULL;
    close->ttd = time(0) + DEFAULT_TTL; /* XXX: Use the server-supplied ttl */
    close->ttd = 0; /* Hack: The root name server IPs never expire */

    /* Add the data to the hash */
    write_lock();
    /* See if the data is already there.  If so, act based on the value
       of if_exists */
    spot_data = mhash_get(dnscache,zone);
    if(spot_data.value != 0 && spot_data.datatype == MARA_DNS_NS) {
        point = spot_data.value;
	switch(if_exists) {
	    case 1: /* Overwrite */
	        /* Destroy the pointer to the element in the hash */
		mhash_undef(dnscache,zone);
		/* Destroy the element itself.  Go down the
		   linked list ans destroy each element one by one,
		   starting at the top. */
	        while(point != 0) {
	           point_save = point->next;
		   if(point->datatype == RR_A && point->data != 0) 
		       js_dealloc(point->data);
                   else if(point->datatype == RR_NS && point->data != 0)
		       js_destroy(point->data);
                   else if(point->data != 0)
		       js_dealloc(point->data);
                   js_dealloc(point);
		   point = point_save;
                   }
                break; 
	    case 2: /* Append */
                while(point->next != 0) 
	            point = point->next; 
                point->next = close;
		write_unlock();
		return JS_SUCCESS;
            /* XXX to do: mask */
	    }
        }
    /* This error condition should never happen */
    else if(spot_data.value != 0) {
        write_unlock();
        return JS_ERROR;
	}
    /* It's a (possibly newly) empty spot--add a new entry to the hash 
       (w/ error checking) */
    else if(mhash_put(dnscache,zone,close,MARA_DNS_NS) == JS_ERROR) {
        write_unlock();
	goto cleanup;
	}
    /* Make sure we do not overflow the cache */
    custodian();
    write_unlock();

    return JS_SUCCESS;
    cleanup_ipp:
        js_dealloc(ipp);
    cleanup:
	js_dealloc(close);
	return JS_ERROR;

    }

/* Add the name of a server "closer" to the answer for a given
   branch of the DNS tree 
   Input:  js_string of zone (raw UDP format with 2-byte suffix), 
           js_string with server name
	   action to take if a node is already here:
	   1) Overwrite 2) Append 3) "Mask" currect table
   Output: JS_ERROR on fatal error, JS_SUCCESS on success
*/

int add_closer_js(js_string *zone, js_string *name, int if_exists) {
    closer *close;
    mhash_e spot_data;
    closer *point, *point_save;
    js_string *new;

    /* Allocate the memory for a new spot on the dnscache hash */
    if((close = js_alloc(1,sizeof(closer))) == 0)
        return JS_ERROR;
    if((new = js_create(name->unit_count + 1,1)) == 0)
        return JS_ERROR;

    if(js_copy(name,new) == JS_ERROR)
        return JS_ERROR;

    /* Make sure the raw zone is of the right rtype */
    if(change_rtype(zone,252) == JS_ERROR)
        goto cleanup;

    /* Fill up the "close" structure */
    close->num_elements = 1;
    close->datatype = RR_NS;
    close->data = new;
    close->next = NULL;
    close->ttd = time(0) + DEFAULT_TTL; /* XXX: Use the server-supplied ttl */

    /* Add the data to the hash */
    write_lock();
    /* See if the data is already there.  If so, add this to the end of
       the chain */
    spot_data = mhash_get(dnscache,zone);

    if(spot_data.value != 0 && spot_data.datatype == MARA_DNS_NS) {
        point = spot_data.value;
	switch(if_exists) {
	    case 1: /* Overwrite */
	        /* Destroy the pointer to the element in the hash */
		mhash_undef(dnscache,zone);
		/* Destroy the element itself.  Go down the
		   linked list ans destroy each element one by one,
		   starting at the top. */
	        while(point != 0) {
	           point_save = point->next;
		   if(point->datatype == RR_A && point->data != 0) 
		       js_dealloc(point->data);
                   else if(point->datatype == RR_NS && point->data != 0)
		       js_destroy(point->data);
                   else if(point->data != 0)
		       js_dealloc(point->data);
                   js_dealloc(point);
		   point = point_save;
                   }
                break; 
	    case 2: /* Append */
		point->num_elements++;
                while(point->next != 0) 
	            point = point->next; 
                point->next = close;
		write_unlock();
		return JS_SUCCESS;
            /* XXX to do: mask */
	    }
        }
    /* This error condition should never happen */
    else if(spot_data.value != 0) {
        write_unlock();
        return JS_ERROR;
        }

    /* It's an empty spot--add a new entry to the hash (w/ error checking) */
    else if(mhash_put(dnscache,zone,close,MARA_DNS_NS) == JS_ERROR) {
        write_unlock();
	goto cleanup;
	}
    /* Make sure we do not overflow the cache */
    custodian();
    write_unlock();

    return JS_SUCCESS;
    cleanup:
        js_destroy(new);
	return JS_ERROR;

    }

/* Wrapper that allows a substring of another js_string object to be
   the zone we add a "closer" jsip for.
   Input: js_string object with desired dlabel as substring
          offset of where the dlabel begins
	  ipjs of IP to add
	  action to do if node is already there
   Output: JS_ERROR on fatal error, JS_SUCCESS on success
*/

int add_closer_jsip_offset(js_string *js, int offset, js_string *ipjs,
                           int if_exists) {

    js_string *sub;
    int len, ret;

    /* Determine how long the domain label is */
    len = dlabel_length(js,offset);
    if(len == JS_ERROR)
        return JS_ERROR;

    /* Make a string which holds just the domain name label */
    if((sub = js_create(len + 4,1)) == 0)
        return JS_ERROR;
    if(js_substr(js,sub,offset,len) == JS_ERROR) {
        js_destroy(sub);
	return JS_ERROR;
	}
    if(js_adduint16(sub,252) == JS_ERROR) {
        js_destroy(sub);
	return JS_ERROR;
	}

    /* Run the actual add_closer_jsip function */
    ret = add_closer_jsip(sub,ipjs,if_exists);
    js_destroy(sub);
    return ret;
    }

/* Wrapper that allows a substring of another js_string object to be
   the zone we add a "closer" js for.
   Input: js_string object with desired dlabel as substring
          offset of where the dlabel (question being answered) begins
	  offset of where the answer itself begins (in same js object)
	  action to do if node is already there
   Output: JS_ERROR on fatal error, JS_SUCCESS on success
*/

int add_closer_js_offset(js_string *js, int offset1, int offset2,
                         int if_exists) {

    js_string *sub1, *sub2;
    int len1, len2, ret;

    /* Determine how long the labels are */
    len1 = dlabel_length(js,offset1);
    if(len1 == JS_ERROR)
        return JS_ERROR;
    len2 = dlabel_length(js,offset2);
    if(len2 == JS_ERROR)
        return JS_ERROR;

    /* Make a string which holds just the question */
    if((sub1 = js_create(len1 + 4,1)) == 0)
        return JS_ERROR;
    if(js_substr(js,sub1,offset1,len1) == JS_ERROR) {
        js_destroy(sub1);
	return JS_ERROR;
	}
    if(js_adduint16(sub1,252) == JS_ERROR) {
        js_destroy(sub1);
	return JS_ERROR;
	}

    /* Make a string which holds just the answer */
    if((sub2 = js_create(len2 + 4,1)) == 0) {
        js_destroy(sub1);
        return JS_ERROR;
        }
    if(js_substr(js,sub2,offset2,len2) == JS_ERROR) {
        js_destroy(sub1);
        js_destroy(sub2);
	return JS_ERROR;
	}
    if(js_adduint16(sub2,1) == JS_ERROR) {
        js_destroy(sub1);
        js_destroy(sub2);
	return JS_ERROR;
	}
     
    /* Run the actual add_closer_js function */
    ret = add_closer_js(sub1,sub2,if_exists);
    js_destroy(sub1);
    js_destroy(sub2);
    return ret;
    }

/* Add an IP as a server "closer" to the answer for a given
   branch of the DNS tree 
   Input:  js_string of zone (raw UDP format with 2-byte suffix), 
           js_string with ip
	   action to take if a node is already here:
	   1) Overwrite 2) Append 3) "Mask" currect table
   Output: JS_ERROR on fatal error, JS_SUCCESS on success
*/

int add_closer_jsip(js_string *zone, js_string *ipjs, int if_exists) {
    closer *close;
    uint32 ip, *ipp;
    closer *point, *point_save;
    mhash_e spot_data;

    /* Allocate the memory for a new spot on the dnscache hash */
    if((close = js_alloc(1,sizeof(closer))) == 0)
        return JS_ERROR;

    /* Make sure the raw zone is of the right rtype */
    if(change_rtype(zone,252) == JS_ERROR)
        goto cleanup;

    /* Convert ipjs in to a format that sendto() can use */
    if(ipjs->unit_count < 4)
        goto cleanup;
    ip = (*(ipjs->string) << 24) | (*(ipjs->string + 1) << 16) | 
         (*(ipjs->string + 2) << 8) | *(ipjs->string + 3);

    /* Fill up the "close" structure */
    close->num_elements = 1;
    close->datatype = RR_A;
    if((ipp = js_alloc(1,sizeof(uint32))) == 0)
        goto cleanup;
    *ipp = ip;
    close->data = ipp;
    close->next = NULL;
    close->ttd = time(0) + DEFAULT_TTL; /* XXX: Use the server-supplied ttl */

    /* Add the data to the hash */
    write_lock();
    /* See if the data is already there.  If so, add this to the end of
       the chain */
    spot_data = mhash_get(dnscache,zone);

    if(spot_data.value != 0 && spot_data.datatype == MARA_DNS_NS) {
        point = spot_data.value;
	switch(if_exists) {
	    case 1: /* Overwrite */
	        /* Destroy the pointer to the element in the hash */
		mhash_undef(dnscache,zone);
		/* Destroy the element itself.  Go down the
		   linked list ans destroy each element one by one,
		   starting at the top. */
	        while(point != 0) {
	           point_save = point->next;
		   if(point->datatype == RR_A && point->data != 0) 
		       js_dealloc(point->data);
                   else if(point->datatype == RR_NS && point->data != 0)
		       js_destroy(point->data);
                   else if(point->data != 0)
		       js_dealloc(point->data);
                   js_dealloc(point);
		   point = point_save;
                   }
                break; 
	    case 2: /* Append */
	        point->num_elements++;
                while(point->next != 0) 
	            point = point->next; 
                point->next = close;
		write_unlock();
		return JS_SUCCESS;
            /* XXX to do: mask */
	    }
        }
    /* This error condition should never happen */
    else if(spot_data.value != 0) {
        write_unlock();
        return JS_ERROR;
        }

    /* It's an empty spot--add a new entry to the hash (w/ error checking) */
    else if(mhash_put(dnscache,zone,close,MARA_DNS_NS) == JS_ERROR) {
        write_unlock();
	goto cleanup;
	}
    /* Make sure we do not fill up the cache */
    custodian();
    write_unlock();

    return JS_SUCCESS;
    cleanup_ipp:
        js_dealloc(ipp);
    cleanup:
        js_destroy(ipjs);
	js_dealloc(close);
	return JS_ERROR;

    }

/* Add an IP as a server "closer" to the answer for a given
   branch of the DNS tree 
   Input:  null-terminated string of zone, nt-string of ddip
   Output: JS_ERROR on fatal error, JS_SUCCESS on success
*/

int add_closer_ip(char *zone, char *ddip) {
    js_string *zone_js, *ddip_js;
    int zone_len, ddip_len, ret;

    /* Sanity checks */
    if(zone == 0 || *zone == 0)
        return JS_ERROR;
    if(ddip == 0 || *ddip == 0)
        return JS_ERROR;

    zone_len = strlen(zone);
    ddip_len = strlen(ddip);

    /* Initialize the js string objects */
    if((zone_js = js_create(zone_len + 3,1)) == 0)
        return JS_ERROR;
 
    if((ddip_js = js_create(ddip_len + 1,1)) == 0) {
        js_destroy(zone_js);
        return JS_ERROR; 
	}

    /* Copy over the null-terminated strings to the js string objects */
    if(js_str2js(zone_js,zone,zone_len,1) == JS_ERROR)
        goto cleanup;
    if(js_str2js(ddip_js,ddip,ddip_len,1) == JS_ERROR)
        goto cleanup;

    /* Convert the zone name in to a binary string */
    if(hname_2rfc1035(zone_js) <= 0)
        goto cleanup;
    /* Add a two-byte "query type" to the end of "zone_js" */
    if(js_adduint16(zone_js,252) == JS_ERROR)
        goto cleanup;
 
    /* Call the "real" function that uses the js_string objects */
    ret = add_closer_jsddip(zone_js,ddip_js,OVERWRITE);
    js_destroy(zone_js);
    js_destroy(ddip_js);
    return ret;

    cleanup:
        js_destroy(ddip_js);
        js_destroy(zone_js);
        return JS_ERROR; 
    }

/* Initialize the cache hash for general use
   Input: maximum number of elements the cache can have
   Output: JS_ERROR on fatal error, JS_SUCCESS on success 
   Global variables used: max_cache, etc.
*/

int init_cache(int max_cache_elements) { 
    closer root;
    js_string *rootns;
    js_string *rootns_zone;
    ipv4pair root_servers[512];
    int bits = 1;
    int exp_counter = 1;
    int counter;
   
    /* Initialize the list of root servers */
    for(counter=0;counter<511;counter++)
        root_servers[counter].ip = 0xffffffff;

    /* Determine the number of bits the hash will need */
    while(exp_counter <= max_cache_elements) {
	/* The enterprise version will come when someone contracts me
	   to write the code (the license has to be public domain or GPL) */
	if(bits > 30)
	    return JS_ERROR;
	bits++;
	exp_counter *= 2;
        }

    /* As a table:
       Max_cache_elements Bits
       0                  1 (hash size 1)
       1                  2 (size 3)
       2-3                3 (" "  7)
       4-7                4 (15)
       8-15               5 (31)
       16-31              6 (63)
       32-63              7 (127)
       64-127             8 (255)
       128-255            9 (511)
       256-511            10 (1023)
       512-1023           11 (2047)
       1024-2047          12 (4095)
       etc.
    */

    if(dnscache != 0) /* We can only init the DNS cache once */
        return JS_ERROR;

    /* Initialize the dnscache object */
    dnscache = (mhash *)mhash_create(bits); /* 1024+ usable elements, 4096
                                             total hash table spots */

    /* Copy over max_cache_elements to a global variable, so that
       other routines can see this value */
    cache_max = max_cache_elements;

    /* Set up the root nameserver.  This is currently very simplistic and
       can only accept a single IP as the root name server.  However, it
       beats me having to constantly repatch the source. */

    if((rootns = js_create(256,1)) == 0)
	return JS_ERROR;
    if((rootns_zone = js_create(256,1)) == 0) {
	js_destroy(rootns);
	return JS_ERROR;
        }

    if(js_qstr2js(rootns,"root_servers") == JS_ERROR) 
	goto cleanup;
    if(js_qstr2js(rootns_zone,".") == JS_ERROR) 
	goto cleanup;
    /* Set the locale so it works with read_dvar */
    js_set_encode(rootns,MARA_LOCALE);
    js_set_encode(rootns_zone,MARA_LOCALE);
    if(read_dvar(rootns,rootns_zone,rootns) == JS_ERROR) 
	goto cleanup;
    /* Reset the locale so it works with the hash routines */
    js_set_encode(rootns,1);
    js_set_encode(rootns_zone,1);
    /* Convert rootns_zone in to a binary packet with "query type" */
    if(js_qstr2js(rootns_zone,"A.") == JS_ERROR) 
	goto cleanup;
    if(hname_2rfc1035(rootns_zone) <= 0) 
	goto cleanup;
    if(js_adduint16(rootns_zone,252) == JS_ERROR)
	goto cleanup;
    /* Convert the rootns in to a list of ipv4 pairs */
    if(make_ip_acl(rootns,root_servers,500,0) == JS_ERROR)
        goto cleanup;
    counter = 0;
    while(root_servers[counter].ip != 0xffffffff) {
        if(counter > 502)
	    goto cleanup;
        if(add_closer_ipv4pair(rootns_zone,&(root_servers[counter]),APPEND) 
           == JS_ERROR)
	    goto cleanup;
        counter++;
        }

#ifdef DEBUG
    debug_explore_dict(dnscache);
#endif

    return JS_SUCCESS;
    cleanup:
	js_destroy(rootns);
	js_destroy(rootns_zone);
	return JS_ERROR;
    }

/* Initialize the cryptographically secure psudo-random-number generator */
/* Input: pointer to string that has filename with binary crypto key */

int init_crypto(js_string *key) {
    unsigned char crypto_key[34];
    int desc;
    char path[MAXPATHLEN];
    /* Initialize the input block and the "binKey" (is this used?) */
    memset(r_inBlock,0,16);
    time((time_t *)&r_inBlock[0]);
    memset(r_binKey,0,16);
    /* Read the key in from the file */
    if(js_js2str(key,path,MAXPATHLEN) == JS_ERROR) {
       return JS_ERROR;
    }
    desc = open(path,O_RDONLY); 
    if(desc == -1)
	return JS_ERROR;
    if(read(desc,crypto_key,16) != 16) /* 16 bytes: 128-bit key */
	return JS_ERROR;
    close(desc);
    if(makeKey(&r_keyInst, DIR_ENCRYPT, 128, crypto_key) != 1)
	return JS_ERROR;
    if(cipherInit(&r_cipherInst, MODE_ECB, NULL) != 1)
	return JS_ERROR;
    if(blockEncrypt(&r_cipherInst,&r_keyInst,r_inBlock,128,r_outBlock) != 128)
	return JS_ERROR;
    
    return JS_SUCCESS;
    }

